/*
 * Copyright (c) 2016-2017, 2019 The Linux Foundation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <asm/io.h>
#include <common.h>
#include <asm/types.h>
#include <fdtdec.h>
#include <asm/arch-qca-common/gpio.h>

DECLARE_GLOBAL_DATA_PTR;

/***********************************************************
* Function description: configure GPIO functinality
* Arguments :
* struct qca_gpio_config gpio_config - GPIO Configuration bits
*
* Return : None
************************************************************/
void gpio_tlmm_config(struct qca_gpio_config *gpio_config)
{
	unsigned int cfg = 0;
	unsigned int val = 0;
	unsigned int *addr = NULL;
	cfg |= gpio_config->pull;
	cfg |= gpio_config->func << 2;
	cfg |= gpio_config->drvstr << 6;
	cfg |= gpio_config->oe << 9;
#ifdef CONFIG_IPQ5018
	cfg |= gpio_config->od_en << 10;
	cfg |= gpio_config->sr_en << 11;
	cfg |= gpio_config->pu_res << 12;
#else
	cfg |= gpio_config->vm << 11;
	cfg |= gpio_config->od_en << 12;
	cfg |= gpio_config->pu_res << 13;
#endif
	if (gpio_config->func == 0) {
		addr = (unsigned int *)GPIO_IN_OUT_ADDR(gpio_config->gpio);
		val = readl(addr);
		val |= gpio_config->out << 1;
		writel(val, addr);
	}

	addr = (unsigned int *)GPIO_CONFIG_ADDR(gpio_config->gpio);
	writel(cfg, addr);

	return;
}

void gpio_tlmm_get_config(struct qca_gpio_config *gpio_config)
{
	unsigned int *addr =(unsigned int *)GPIO_CONFIG_ADDR(gpio_config->gpio);

	unsigned int val = readl(addr);

	gpio_config->pull = val & 0x03;
	gpio_config->func = (val >> 2) & 0x07;
	gpio_config->drvstr = (val >> 6) & 0x07;
	gpio_config->oe = (val >> 9) & 0x03;
	gpio_config->vm = (val >> 11) & 0x01;
	gpio_config->od_en = (val >> 12) & 0x01;
	gpio_config->pu_res = (val >> 13) & 0x01;

}
void gpio_set_value(unsigned int gpio, unsigned int out)
{
	unsigned int *addr = (unsigned int *)GPIO_IN_OUT_ADDR(gpio);
	unsigned int val = 0;


	struct qca_gpio_config gpio_config;
	gpio_config.gpio = gpio;
	gpio_tlmm_get_config(&gpio_config);
	if (gpio_config.func != 0) {
		printf("Failed to set gpio %d to value %d because it isn't in GPIO, func=%d\n", gpio, out, gpio_config.func);
		return;
	}
	val = readl(addr);
	val &= ~(0x2);
	val |= out << 1;
	writel(val, addr);

	if (gpio_config.oe == 0) {
		gpio_config.oe = 1;
		gpio_config.out = out;
		gpio_tlmm_config(&gpio_config);
	}
}
void get_gpio_config(void)
{
	for (int i = 0; i < 69; i++) {
		struct qca_gpio_config gpio_config;
		memset(&gpio_config, 0, sizeof(struct qca_gpio_config));
		gpio_config.gpio = i;
		gpio_tlmm_get_config(&gpio_config);
		print_gpio_config(&gpio_config);
	}
}
int gpio_get_value(unsigned int gpio)
{
	unsigned int *addr = (unsigned int *)GPIO_IN_OUT_ADDR(gpio);
	unsigned int val = readl(addr);

	return (val & 1);
}

int qca_gpio_get_value(struct qca_gpio_config * gpio)
{
	int value = !!gpio_get_value(gpio->gpio);
	if (gpio->active_low)
		return !value;
	return value;
}

void qca_gpio_set_value(struct qca_gpio_config * gpio, int value)
{
	value = !!value;
	if (gpio->active_low){
		value = !value;
	}
	gpio_set_value(gpio->gpio, value);
}
void gpio_toggle_value(unsigned int gpio)
{
	int value = gpio_get_value(gpio);
	printf("GPIO %d value was %d", gpio, value);
	gpio_set_value(gpio, !!!value); /* ! */
	printf(" and is now %d\n", gpio_get_value(gpio));
}
void gpio_direction_output(unsigned int gpio, unsigned int out)
{
	unsigned int *addr = (unsigned int *)GPIO_CONFIG_ADDR(gpio);
	unsigned int val = 0;

	gpio_set_value(gpio, out);
	val = readl(addr);
	val |= 1 << 9;
	writel(val, addr);
}

void print_gpio_config(struct qca_gpio_config* config)
{

	printf("GPIO_%d: %d {func=%d,output=%d,pull=%d,drvstr=%d,oe=%d,vm=%d,od_en=%d,pu_res=%d}\n",
			config->gpio, gpio_get_value(config->gpio),
			config->func, config->out, config->pull, config->drvstr,
			config->oe, config->vm, config->od_en, config->pu_res);

}

struct qca_gpio_config * fdt_qca_gpio(int offset) {

	struct qca_gpio_config * gpio_config = (struct qca_gpio_config *) malloc(sizeof(struct qca_gpio_config));
	if (!gpio_config) {
		printf("Warning, failed to allocate the qca_gpio_config\n");
		return NULL;
	}
	gpio_config->gpio	= fdtdec_get_uint(gd->fdt_blob, offset, "gpio", 0);
	gpio_config->func	= fdtdec_get_uint(gd->fdt_blob, offset, "func", 0);
	gpio_config->out		= fdtdec_get_uint(gd->fdt_blob, offset, "out", 0);
	gpio_config->pull	= fdtdec_get_uint(gd->fdt_blob, offset, "pull", 0);
	gpio_config->drvstr	= fdtdec_get_uint(gd->fdt_blob, offset, "drvstr", 0);
	gpio_config->oe		= fdtdec_get_uint(gd->fdt_blob, offset, "oe", 0);
	gpio_config->vm		= fdtdec_get_uint(gd->fdt_blob, offset, "vm", 0);
	gpio_config->od_en	= fdtdec_get_uint(gd->fdt_blob, offset, "od_en", 0);
	gpio_config->pu_res	= fdtdec_get_uint(gd->fdt_blob, offset, "pu_res", 0);
	gpio_config->pu_res	= fdtdec_get_uint(gd->fdt_blob, offset, "pu_res", 0);
	gpio_config->active_low = fdtdec_get_uint(gd->fdt_blob, offset, "active-low", 0);
	return gpio_config;
}
int qca_gpio_init(int offset)
{
	struct qca_gpio_config gpio_config;

	for (offset = fdt_first_subnode(gd->fdt_blob, offset); offset > 0;
	     offset = fdt_next_subnode(gd->fdt_blob, offset)) {

		memset(&gpio_config, 0, sizeof(struct qca_gpio_config));
		gpio_config.gpio	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "gpio", 0);
		gpio_config.func	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "func", 0);
		gpio_config.out		= fdtdec_get_uint(gd->fdt_blob,
							  offset, "out", 0);
		gpio_config.pull	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "pull", 0);
		gpio_config.drvstr	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "drvstr", 0);
		gpio_config.oe		= fdtdec_get_uint(gd->fdt_blob,
							  offset, "oe", 0);
		gpio_config.vm		= fdtdec_get_uint(gd->fdt_blob,
							  offset, "vm", 0);
		gpio_config.od_en	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "od_en", 0);
		gpio_config.pu_res	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "pu_res", 0);
		gpio_config.sr_en	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "sr_en", 0);
		gpio_tlmm_config(&gpio_config);
	}
	return 0;
}

int qca_gpio_deinit(int offset)
{
	struct qca_gpio_config gpio_config;

	for (offset = fdt_first_subnode(gd->fdt_blob, offset); offset > 0;
	     offset = fdt_next_subnode(gd->fdt_blob, offset)) {

		gpio_config.gpio	= fdtdec_get_uint(gd->fdt_blob,
							  offset, "gpio", 0);
		unsigned int *addr =
			(unsigned int *)GPIO_CONFIG_ADDR(gpio_config.gpio);
		writel(1, addr);
		addr = (unsigned int *)GPIO_IN_OUT_ADDR(gpio_config.gpio);
		writel(1, addr);
	}
	return 0;
}
